# TODO recode cause too complex

# TODO --> clean and reimplement this properly...
# alternatively offer commands to be added to the figure so that they fit the requirements --> is another way of doing

# check if that is interesting for me https://matplotlib.org/stable/devel/MEP/MEP25.html
# CROP IS NOT OPTIMAL BUT OK FOR NOW...

# this file must be a valid svg file and/or a matplotlib figure --> TODO handle both and check that everything correct

# see also pyQT_svg_drawing_test.py --> for loading and reading svg files --> maybe this is what I should do ???

# TODO make it draw as svg rather than as raster directly --> see how
# in fact only required for save

# le retaillage des graphes a l'air de marcher mais à verif qd meme...
# nb the drawing as svg is only required for saving and not for displaying

# maybe add behaviour should be different if images are missing --> then in that case the images should be added to the missing positions, this is important for panels
# maybe instead of panel I can build complex layouts and then fill them by adding images to them
# maybe also support for human language (not sure how easy that would be), maybe return a list of commands

# TODO see my plots_python_test.py

# a good start but I really need to have handle the size better and the resize too --> think how to smartly do that ???

# maybe ask if keep AR should be set to true ??? or not think about it...
# https://docs.python.org/2/library/operator.html

# TODO may also contain svg or graphs or ???
# TODO handle extra labels for images directly and also for rows or cols --> think how to do that but must be doable

# I probably need python fig and axes
# https://matplotlib.org/3.2.2/gallery/subplots_axes_and_figures/gridspec_multicolumn.html#sphx-glr-gallery-subplots-axes-and-figures-gridspec-multicolumn-py
# https://matplotlib.org/3.2.2/api/_as_gen/matplotlib.pyplot.figure.html
# https://matplotlib.org/3.2.2/api/matplotlib_configuration_api.html#matplotlib.rcParams
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import traceback

from qtpy.QtSvg import QSvgRenderer

from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.draw.shapes.txt2d import TAText2D
from epyseg.img import Img
from qtpy.QtCore import QRectF
import base64
from PIL import Image
import matplotlib.pyplot as plt
import io
import numpy as np
# logger
from epyseg.tools.logger import TA_logger

# should this stuff rather implement Image2D --> maybe or not think about it...

logger = TA_logger()


# do I really need crop for that and shouldn't it be done from outside to prevent pbs and it will also definitely mess with the font which is not a good thing ???
# try it

# I don't manage to get the crops to work...
# can I add a hashtag to texts so that all texts with the same hashtag are treated the same...

class VectorGraphics2D(Rect2D):  # Image2D # Rect2D

    def __init__(self, figure, *args, x=None, y=None, width=None, height=None, opacity=1., stroke=0.65, **kwargs):
        self.isSet = False

        # TODO can only force size of a graph but not of an svg --> modify code accordingly, if graph and modify need apply set_size_inches to the figure... --> TODO rapidly
        tmp_size = QRectF()
        if x is not None and str(x).isnumeric():
            tmp_size.setX(x)
        if y is not None and str(x).isnumeric():
            tmp_size.setY(y)
        if width is not None and str(width).isnumeric():
            tmp_size.setWidth(width)
        if height is not None and str(height).isnumeric():
            tmp_size.setHeight(height)
        super(VectorGraphics2D, self).__init__(tmp_size)

        # self.annotation = [] # should contain the objects for annotating imaging --> shapes and texts
        self.letter = None  # when objects are swapped need change the letter
        self.figure = None
        self.annotation = []  # should contain the objects for annotating imaging --> shapes and texts
        # self.img = None
        # self.qimage = None
        self.renderer = None
        self.filename = None
        # crops
        self.__crop_left = None
        self.__crop_right = None
        self.__crop_top = None
        self.__crop_bottom = None
        self.must_update_figure_on_first_paint = False

        # first argument should be fig
        # if args:
        #     if len(args) == 1:
        #         self.filename = args[0]
        # else:
        #     self.filename = None

        # if x is None and y is None and width is not None and height is not None:
        #     super(Graph2D, self).__init__(0, 0, width, height)
        #     self.isSet = True
        # elif x is None and y is None and width is None and height is None:
        #     # print('in 0')
        #     self.img = Img(self.filename)
        #     self.qimage = self.img.getQimage()
        #     width = self.img.get_width()
        #     height = self.img.get_height()
        #     super(Graph2D, self).__init__(0, 0, width, height)
        #     self.isSet = True
        # elif x is not None and y is not None and width is not None and height is not None:
        #     self.img = None
        #     super(Graph2D, self).__init__(x, y, width, height)
        #     self.isSet = True
        # elif data is None:
        #     if self.filename is not None:
        #         self.img = Img(self.filename)
        #         self.qimage = self.img.getQimage()
        #         if x is None:
        #             x = 0
        #         if y is None:
        #             y = 0
        #         super(Graph2D, self).__init__(x, y, self.img.get_width(), self.img.get_height())
        #         self.isSet = True
        # elif data is not None:
        #     self.img = Img(data,
        #                    dimensions=dimensions)  # need width and height so cannot really be only a numpy stuff --> cause no width or height by default --> or need tags such as image type for dimensions
        #     self.qimage = self.img.getQimage()
        #     # need Image dimensions id data is not of type IMG --> could check that
        #     if x is None:
        #         x = 0
        #     if y is None:
        #         y = 0
        #     super(Graph2D, self).__init__(x, y, self.img.get_width(), self.img.get_height())
        # TODO should I also allow aspect ratio or keep aspect ratio or fixed width or height or auto --> think about it and simply try

        # TODO should I allow to create a fig from some code using eval (i.e. a script that can be serialized, given the dangers of the eval function I dunno)
        # if self.figure is not None and isinstance(self.figure, plt.figure):
        #     self.isSet = True
        # else:
        #     self.isSet = False
        self.setFigure(figure)
        # super(Graph2D, self).__init__(0, 0, self.img.get_width(), self.img.get_height())

        # store also the image for that

        self.stroke = stroke  # DO I REALLY NEED STROKE
        self.opacity = opacity

    def getFigure(self):
        return self.figure

    def setFigure(self, figure):

        self.must_update_figure_on_first_paint = False

        # TODO load the raw image there so that it can be drawn easily
        # self.figure = figure
        if figure is not None and isinstance(figure, plt.Figure):

            # self.img = self.toImg()
            # print(self.img.get_width())
            # self.qimage = self.img.getQimage()
            # make a renderer out of it and display it ...

            # print("size inches before rendering", figure.get_size_inches())
            self.figure = figure
            buffer = self._toBuffer(bufferType='svg')
            self.renderer = QSvgRenderer(buffer.read())
            buffer.close()
            # self.setSize(QSizeF(self.renderer.defaultSize()))
            # upon init we do set the width --> should this be done here or at other position ??? think about it
            if self.width() == 0:
                size = self.renderer.defaultSize()
                # print('default size', size, 'vs', self.renderer.viewBox())
                self.setWidth(size.width())
                self.setHeight(size.height())
            self.isSet = True
        elif figure is not None and isinstance(figure, str):  # path to an svg file
            # just try load it
            # do I ever need the buffer for this too cause if yes then I would need to get it --> think how ???

            # if buffer is needed --> this is how I should do it
            # in_file = open(figure)  # opening for [r]eading as [b]inary
            # data = in_file.read()  # if you only wanted to read 512 bytes, do .read(512)
            # in_file.close()
            # TODO add tries to see if that works and if opened properly
            # print(data)
            self.renderer = QSvgRenderer(figure)  # data --> need convert data to be able to read it

            # self.renderer.render(painter, self)# the stuff is a qrectf so that should work

            # self.renderer.setViewBox(viewbox)

            self.filename = figure
            if self.width() == 0:
                size = self.renderer.defaultSize()
                # print('default size', size, 'vs', self.renderer.viewBox())
                self.setWidth(size.width())
                self.setHeight(size.height())
                if size.width() <= 0:
                    logger.error('image "' + str(self.filename) + '" could not be loaded')
                    self.isSet = False
                    return
                # j'arrive pas à faire des crop avec les viewbox
                # viewbox = self.renderer.viewBoxF()
                #
                # # does not really work
                # neo = QRectF()
                # neo.setX(30)
                # neo.setY(30)
                # neo.setHeight(self.height()-30)
                # neo.setWidth(self.width()-30)
                # self.renderer.setViewBox(neo)

            self.figure = None
            self.isSet = True
        else:
            logger.error(
                'The provided figure is not a valid matplotlib figure nor a valid svg file! Nothing can be done with it... Sorry...')
            self.figure = None
            self.isSet = False


    def getAxes(self):
        if self.figure is not None:
            return self.figure.axes
        return None


# somehow I need it --> keep it for now !!!
# WHY should I have that ???
#     @return the block incompressible width
    def getIncompressibleWidth(self):
        extra_space = 0  # can add some if boxes around to add text
        return extra_space

    # @return the block incompressible height
    def getIncompressibleHeight(self):
        extra_space = 0  # can add some if boxes around to add text
        return extra_space

    # if bg color is set then need add it --> see how to do that
    def setLettering(self, letter):
        if isinstance(letter, TAText2D):
            self.letter = letter
        elif isinstance(letter, str):
            if letter.strip() == '':
                self.letter = None
            else:
                self.letter = TAText2D(letter)

    def _toBuffer(self, bufferType='raster'):
        if self.figure is not None:
            buf = io.BytesIO()
            if bufferType == 'raster':
                self.figure.savefig(buf, format='png', bbox_inches='tight')
            else:
                self.figure.savefig(buf, format='svg', bbox_inches='tight')
            buf.seek(0)
            return buf
        return None

    # will only work for figures and is that even needed in fact --> think about it
    def _toBase64(self):
        buf = self._toBuffer()
        if buf is None:
            return None
        buf.seek(0)
        figdata_png = base64.b64encode(buf.getvalue())
        buf.close()
        return figdata_png

    # will only work for figures and is that even needed in fact --> think about it
    def toSVG(self):
        buf = self._toBuffer()
        if buf is None:
            return None
        buf.seek(0)
        text = buf.read()
        buf.close()
        return text

    # will only work for figures and is that even needed in fact --> think about it
    def _toImg(self):
        # print(self.toBase64()) # this is ok
        if self.figure is not None:
            buf = self._toBuffer()
            if buf is None:
                return None
            buf.seek(0)
            im = Image.open(buf)

            # im.show()
            pix = np.array(im)

            # print(pix.shape)
            img = Img(pix, dimensions='hwc')
            # print(img.shape, pix.shape)

            buf.close()
            # should I get image width and height there ???
            # im.show()
            return img
        return None

    # NEED CONVERSION FROM AND TO PX AND CM AND INCHES... --> TODO
    # TODO do a class to convert stuff and do it smart enough to handle any type of data and return the same type e.g. tuples as tuples, numbers as numbers and lists as lists
    # https://stackoverflow.com/questions/14708695/specify-figure-size-in-centimeter-in-matplotlib
    def cm2inch(self, *tupl):
        inch = 2.54
        if isinstance(tupl[0], tuple):
            return tuple(i / inch for i in tupl[0])
        else:
            # print(tupl) # why does it contain a qrectf ????
            return tuple(i / inch for i in tupl)

    def draw(self, painter, draw=True):
        # TODO check if this is really what I want at least it seems better than what I was doing before!!!
        if self.must_update_figure_on_first_paint:
            try:
                if self.figure is not None:
                    cm_width = 0.02646 * self.width()
                    cm_height = 0.02646 * self.height()
                    self.figure.set_size_inches(self.cm2inch(cm_width * self.scale, cm_height * self.scale))  # need do a px to
                    self.setFigure(self.figure)
            except:
                traceback.print_exc()
            self.must_update_figure_on_first_paint = False

        if draw:
            painter.save()
        painter.setOpacity(self.opacity)

        # may also work --> is that the best solution ????
        # painter.scale(0.3, 0.3)

        # new addition but probably not ok
        painter.scale(self.scale, self.scale)  # size is ok but position sucks --> how can I fix that
        # --> ok except for the coords --> fix that



        if draw and self.renderer is not None:
            # if self.img is not None:
            #     qsource = QRectF(0,0,self.img.get_width(), self.img.get_height())
            #     painter.drawImage(self, self.qimage , qsource) # , flags=QtCore.Qt.AutoColor
            # else:
            #     painter.drawRect(self)

            # print('size of drawing', self)

            # print('view box',self.renderer.viewBoxF())

            # viewbox = self.renderer.viewBoxF()

            # --> vraiment presque ça
            # does not really work
            neo = QRectF(self)

            # neo*=self.scale
            # neo.setX(self.x()-10)
            # neo.setY(self.y()-10)
            # neo.setHeight(self.height()+30)
            # neo.setWidth(self.width()+30)

            # print(neo)

            # nb this will create a shear if one dim is not cropped as the other --> really not great in fact maybe deactivate for now ???? or compute AR and adapt it
            # TODO just warn that it's buggy and should not be used for SVG files only, ok for other stuff though

            # can I preserve AR ???

            if self.__crop_left is not None:
                neo.setX(neo.x() - self.__crop_left)
                neo.setWidth(neo.width() + self.__crop_left)
            if self.__crop_top is not None:
                neo.setY(neo.y() - self.__crop_top)
                neo.setHeight(neo.height() + self.__crop_top)
            if self.__crop_bottom is not None:
                neo.setHeight(neo.height() + self.__crop_bottom)
            if self.__crop_right is not None:
                neo.setWidth(neo.width() + self.__crop_right)

            # maintenant ça a l'air bon...

            # --> ça c'est ok --> c'est le clip rect qui merde du coup

            # print('view box neo', neo)
            # self.renderer.setViewBox(neo)
            # le clipping marche mais faut le combiner avec autre chose
            # painter.setClipRect(self.x()+10, self.y()+10, self.width()-30,self.height()-30)#, Qt::ClipOperation operation = Qt::ReplaceClip

            # TODO KEEP UNFORTUNATELY  unfortunately cropping does not work when saved as svg but works when saved as raster... see https://bugreports.qt.io/browse/QTBUG-28636
            # maybe do masque d'ecretage in illustrator or inkscape https://linuxgraphic.org/forums/viewtopic.php?f=6&t=6437
            # TODO KEEP IT PROBABLY ALSO CREATES A ZOOM THAT WILL MESS WITH THE FONTS AND LINE SIZE...
            painter.setClipRect(self)  # , Qt::ClipOperation operation = Qt::ReplaceClip , operation=Qt.ReplaceClip
            # neo.setX(self.x())
            # neo.setY(self.y())


            # print('neo vs self',neo, self) # probably a bug here too

            self.renderer.render(painter, neo)  # the stuff is a qrectf so that should work
            # self.sequence_scene.render(painter, target=neo, source=self)

            painter.restore()
            # self.renderer.setViewBox(viewbox)

        # then need to draw the letter
        if self.letter is not None:
            self.letter.set_P1(self.get_P1())
            self.letter.draw(painter)

        if self.annotation is not None and self.annotation:
            for annot in self.annotation:
                annot.draw(draw=draw)

    # def fill(self, painter, draw=True):
    # if self.fill_color is None:
    #     return
    # if draw:
    #     painter.save()
    # painter.setOpacity(self.opacity)
    # if draw:
    #     if self.img is not None:
    #         qsource = QRectF(0, 0, self.img.get_width(), self.img.get_height())
    #         painter.drawImage(self, self.qimage , qsource)
    #     else:
    #         painter.drawRect(self)
    #     painter.restore()
    # return self.draw(painter=painter, draw=draw)

    # def drawAndFill(self, painter):
    # painter.save()
    # if self.img is not None:
    #     qsource = QRectF(0, 0, self.img.get_width(), self.img.get_height())
    #     painter.drawImage(self, self.qimage , qsource)
    # else:
    #     painter.drawRect(self)
    # painter.restore()
    # return self.draw(painter=painter)

    def __add__(self, other):
        from epyseg.figure.row import Row  # KEEP Really required to avoid circular imports
        return Row(self, other)

    # create a Fig with divide
    def __truediv__(self, other):
        from epyseg.draw.widgets.col import col  # KEEP Really required to avoid circular imports
        return col(self, other)

    # Force the montage width to equal 'width_in_px'

    # ne marche qu'avec les graphes en fait et faudrait demander si on peut changer aspect ratio... avant
    def setToWidth(self, width_in_px):
        pure_image_width = self.width()
        ratio = width_in_px / pure_image_width
        self.setWidth(width_in_px)
        self.setHeight(self.height() * ratio)

        self.scale = width_in_px / pure_image_width

        # nb this is what makes everythong so slow --> skip that and do it only when necessary
        if self.figure is not None:
            # need recompute the graph and update it with the new size --> convert in pixels
            cm_width = 0.02646 * width_in_px
            cm_height = 0.02646 * self.height()

            self.must_update_figure_on_first_paint = True
            # MEGA TODO maybe set a boolean saying that the fig needs be updated on plot, then set it to false when this is done... --> set the figure so that it fits the image
            # see how I can do that ???


            # print('cm', cm_width, cm_height)
            # print('cm', cm_width, cm_height)
            # print(self.figure)
            # print('abs')
            # print('inches', self.cm2inch(cm_width,cm_height)) # bug is here
            # print('ibs')
            # self.figure.set_size_inches(self.cm2inch(cm_width, cm_height))  # need do a px to inch or to cm --> TODO

            # MEGA ULTIMATE TODO freshly inactivated because too slow --> need do it in a smart way only in the end --> this is really the slow part and need only be done on save or export --> see how to do that anyways it does not make sense to have this here
            # self.figure.set_size_inches(self.cm2inch(cm_width*self.scale, cm_height*self.scale))  # need do a px to inch or to cm --> TODO


            # print(self.figure.get_size_inches())
            # print('size before', self) # --> size ok after not ok --> bug in conversion somewhere

            # MEGA ULTIMATE TODO freshly inactivated because too slow --> need do it in a smart way only in the end --> this is really the slow part and need only be done on save or export --> see how to do that anyways it does not make sense to have this here
            # self.setFigure(self.figure)


            # print('changing size')
            # print('inches after', self.figure.get_size_inches())
            # print('size after', self)

    def setToHeight(self, height_in_px):
        pure_image_height = self.height()
        self.setHeight(height_in_px)
        ratio = height_in_px / pure_image_height
        self.setWidth(self.width() * ratio)
        # TODO implement that
        self.scale = height_in_px / pure_image_height

        if self.figure is not None:

            # MEGA TODO maybe set a boolean saying that the fig needs be updated on plot, then set it to false when this is done...

            # the lines below cause a bug I guess it's because it's an image object and not a qrect object anymore and so width has pb --> calls a function
            cm_width = 0.02646 * self.width()
            cm_height = 0.02646 * height_in_px
            self.must_update_figure_on_first_paint = True

            # print('cm', cm_width, cm_height)
            # print(self.figure)
            # print('oubs')
            # print(self.cm2inch(cm_width, cm_height))
            # print('ebs')
            # self.figure.set_size_inches(self.cm2inch(cm_width, cm_height))  # need do a px to inch or to cm --> TODO

            # MEGA ULTIMATE TODO freshly inactivated because too slow --> need do it in a smart way only in the end --> this is really the slow part and need only be done on save or export --> see how to do that anyways it does not make sense to have this here
            # self.figure.set_size_inches(self.cm2inch(cm_width*self.scale, cm_height*self.scale))  # need do a px to inch or to cm --> TODO

            # print(self.figure.get_size_inches())

            # MEGA ULTIMATE TODO freshly inactivated because too slow --> need do it in a smart way only in the end --> this is really the slow part and need only be done on save or export --> see how to do that anyways it does not make sense to have this here
            # self.setFigure(self.figure)

    # No clue how to do that --> ignore...
    def crop(self, left=None, right=None, top=None, bottom=None, all=None):
        logger.warning(
            'Crop of svg files is very buggy and should not be used, especially for scientific publications as it may distort the image...')
        # print(self.boundingRect())
        if left is not None:
            self.__crop_left = left
            self.setWidth(self.width() - self.__crop_left)
            self.setX(self.x() + self.__crop_left)
        if right is not None:
            self.__crop_right = right
            self.setWidth(self.width() - self.__crop_right)
        if top is not None:
            self.__crop_top = top
            self.setHeight(self.height() - self.__crop_top)
            self.setY(self.y() + self.__crop_top)
        if bottom is not None:
            self.__crop_bottom = bottom
            self.setHeight(self.height() - self.__crop_bottom)
        if all is not None:
            self.__crop_left = all
            self.__crop_right = all
            self.__crop_top = all
            self.__crop_bottom = all
            self.setX(self.x() + self.__crop_left)
            self.setY(self.y() + self.__crop_top)
            self.setWidth(self.width() - self.__crop_left)
            self.setWidth(self.width() - self.__crop_right)
            self.setHeight(self.height() - self.__crop_top)
            self.setHeight(self.height() - self.__crop_bottom)

        # see how to crop actually because I need to create a qimage
        # self.qimage = self.img.crop()
        # print(self.boundingRect())


if __name__ == '__main__':
    # Data for plotting
    t = np.arange(0.0, 2.0, 0.01)
    s = 1 + np.sin(2 * np.pi * t)

    fig, ax = plt.subplots()
    ax.plot(t, s)

    ax.set(xlabel='time (s)', ylabel='voltage (mV)',
           title='About as simple as it gets, folks')
    ax.grid()

    # plt.show()
    # fig.savefig("test.png")
    # plt.show()
    # ça marche --> voici deux examples de shapes
    test = VectorGraphics2D(fig, x=12, y=0, width=100, height=100)  # could also be used to create empty image with

    # print(test.img)
    print(test.boundingRect())
    print(test.get_P1().x())

    # bug qd on definit une image comme param
    # test = Image2D('./../data/counter/06.png')
    # test = Graph2D('D:/dataset1/unseen/100708_png06.png')
    print(test.boundingRect())  # --> it is ok there so why not below # not callable --> why -->
    print(test.get_P1())  # ça marche donc où est le bug
    print(test.get_P1().y())  # ça marche donc où est le bug
    # print(test.getP1().width())
    # ça marche

    test.setToWidth(512)
    print(test.boundingRect())
    # --> ok
    # how can I think the scale to that ???

    test2 = VectorGraphics2D('/E/Sample_images/sample_images_svg/cartman.svg', x=12, y=0, width=100, height=100)



# really good it did fix most of the errors --> try fix the little remaining errors and see if things can be sped up such as by passing vertices as they should not change from run to run and it is the slow part of the code --> like that could do tremendous amounts of recursion
# TODO really try to fix the tracks before saving when using the other code --> see how well it works when integrated to the other
# could also use this to fix the translations by smthg more realistic --> try it maybe --> may also be used to further fix swapping issues maybe --> give it a try!!!

# seems really good now but really need do it recursively... --> give it a try


# bug found I do miss plenty of vertices --> try fix that


# do smthg to detect swaps and also find a way to fix them
# finalize my robot so that I can save and do better things --> finalize my robot
# for each cell detect its neighbor in all cases and then try a match --> if they don't keep half of their neighbors there is a swap --> then try fix it pb --> need be done recursively

# try several matches for the region of interest


# load an image then try to have it to
from epyseg.img import int24_to_RGB, RGB_to_int24
from timeit import default_timer as timer
import numpy as np
from epyseg.img import Img
# really very good --> now make it recursive then done --> extract the cell from there and continue
from natsort import natsorted
import os

# parameters
from epyseg.utils.loadlist import loadlist
from epyseg.ta.tracking.rapid_detection_of_vertices import associate_vertices_to_cells, detect_vertices_and_bonds

__DEBUG__ = False


# do the saving outside and just pass images as they should be passed in there
# TODO pass vertices here to gain time and reassign neighbors to vertices
def _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1, vertices_t0=None, vertices_t1=None, MIN_CUT_OFF_SWAPPING=None):
    if __DEBUG__:
        cell_to_check = 0x29d785  # 0x976494 #0x29d785  # 0x746502 #0x96c9ef #0x542358  # 0xf36619 #0x496bbd #0x1bf38f
        # map           9921684   2742149   7628034  9882095   5514072   15951385   4811709
        print('cell to check', cell_to_check, format(cell_to_check, 'x'))

    # --> marche --> sauve pas en composite --> en fait c'est ce que je veux dans certains cas --> void si je peux pas hacker un peu mon truc --> si 3 channels --> mettre en RGB or put a force composite of force non composite
    # import tifffile
    # tifffile.imwrite(os.path.join(filename0_without_ext,'tra_test_saving_24bits_0.tif'), tracked_cells_t0, imagej=True,                      metadata={})
    # tifffile.imwrite(os.path.join(filename0_without_ext,'tra_test_saving_24bits_0.tif'), tracked_cells_t0) # ça aussi ça marche en fait --> could maybe put software inside to avoid issues

    # plt.imshow(tracked_cells_t0)
    # plt.show()

    # now detect vertices and neighbs for each cell and see which is what
    # need convert the image to a single RGB image

    # should be done now
    # tracked_cells_t0 = RGB_to_int24(tracked_cells_t0)
    # tracked_cells_t1 = RGB_to_int24(tracked_cells_t1)

    # print('current time', timer()-inner_time) #

    # Img(int24_to_RGB(tracked_cells_t0)).save(os.path.join(filename0_without_ext, 'tra_test_saving_24bits.tif'),
    #                                             mode='raw')
    # import tifffile
    # tifffile.imwrite(os.path.join(filename0_without_ext,'tra_test_saving_24bits.tif'), tracked_cells_t0, imagej=False,                      metadata={})

    # TODO need resplit to save as IJ --> TODO

    # tifffile.imwrite(os.path.join(filename0_without_ext,'tra_test_saving_24bits.tif'), tracked_cells_t0.astype(np.int32)) # marche pas tout à fait car l'image est convertie en float --> see how to fix that
    # tifffile.imsave(tracked_cells_t0, photometric='rgb')
    # Img(tracked_cells_t0.astype(np.int32)).save(os.path.join(filename0_without_ext,'tra_test_saving_24bits.tif'))

    # now TODO just track vertices and alike --> TODO
    # t0_cells_n_vertices = {}
    # t1_cells_n_vertices = {}

    # NB could also be done using the cell neighbors centroids --> that would really help I think and could allow the matching of several cells together --> so maybe really worth giving it a try --> the neighbor mapping could also help me map things better anyway
    # nb --> this increases time by 3
    # NOW WE ASSOCIATE VERTICES TO EVERY CELL
    # detect_vertices_and_bonds

    # inner_time = timer()
    # print(detect_vertices_and_bonds(tracked_cells_t0)) # --> an image with vertices --> is the slow part --> speed it up
    # print('current time', timer() - inner_time)

    # inner_time = timer()
    if vertices_t0 is None:
        vertices_t0 = detect_vertices_and_bonds(tracked_cells_t0)
        vertices_t0 = np.where(vertices_t0 == 255)
    if vertices_t1 is None:
        vertices_t1 = detect_vertices_and_bonds(tracked_cells_t1)
        vertices_t1 = np.where(vertices_t1 == 255)
    # print('current time0', timer() - inner_time)

    # inner_time = timer()
    t0_cells_n_vertices = associate_vertices_to_cells(tracked_cells_t0, vertices_t0,
                                                      forbidden_colors=(0, 0xFFFFFF),
                                                      output_cells_and_their_neighbors=True)
    t1_cells_n_vertices = associate_vertices_to_cells(tracked_cells_t1, vertices_t1,
                                                      forbidden_colors=(0, 0xFFFFFF),
                                                      output_cells_and_their_neighbors=True)
    # print('current time', timer() - inner_time)  # now most of the time is taken there --> speed it up then done
    # print('current time', timer() - inner_time) # environ 1/3 du temps --> peut etre accelere je pense

    # for debug
    # Img(detect_vertices_and_bonds(tracked_cells_t1).astype(np.uint8)).save(
    #     os.path.join(filename1_without_ext, 'tra_vertices.tif'))

    # au lieu des vertices je veux les IDs des neighbs --> comment les recuperer --> à voir en fait...

    if __DEBUG__:
        print('t0_cells_n_vertices', t0_cells_n_vertices)
        print('t1_cells_n_vertices', t1_cells_n_vertices)

    # print('end loop', timer() - start_loop)

    # now effectively detect all swaps --> TODO
    '''
        # old code swap detection
        # swapped_cells = []
        # for key, value in t1_cells_n_vertices.items():
        #     # print(key, '->', value)
        #     cell_id_t1 = key
        #     if cell_id_t1 in t0_cells_n_vertices:
        #         # cell exists in previous image --> see if swapped...
        #         neighs_at_t0 = t0_cells_n_vertices[cell_id_t1]
        #         neighs_at_t1 = t1_cells_n_vertices[cell_id_t1]
        #         common_neighs = set(neighs_at_t1).intersection(neighs_at_t0)
        # 
        #         if len(common_neighs) < len(neighs_at_t1) / 2:
        #             if __DEBUG__:
        #                 print('swapped cell', cell_id_t1, cell_id_t1 >> 16 & 0xFF, cell_id_t1 >> 8 & 0xFF,
        #                       cell_id_t1 & 0xFF)
        #                 print('common_neighs', len(common_neighs), len(neighs_at_t0), len(neighs_at_t1), common_neighs)
        #                 print('real neighbs', neighs_at_t0, neighs_at_t1)
        # 
        #             swapped_cells.append(cell_id_t1)
        # 
        # # second step do a matcher in order to perform
        # # if stuck for 10 secs after reboot --> launch a rescue
        # # would be smarter to read the temp files rather than do nothing
        # # try to set an alarm
        # # or do a user overrun that executes step by step what I tell it to do --> could use and broadcast commands and snaps
        # 
        # swapped_cells_image = np.zeros_like(tracked_cells_t0)
        # 
        # for id in swapped_cells:
        #     swapped_cells_image[tracked_cells_t1 == id] = id
        # 
        # if __DEBUG__:
        #     print(swapped_cells_image.dtype)
        # 
        #     # Img(int24_to_RGB(swapped_cells_image)).save(os.path.join(filename0_without_ext, 'track_swapped_cells_only.tif'), mode='raw')
        # 
        #     # I MAY HAVE A BUG IN THE CODE BELOW BECAUSE THE OTHER CODE IS MUCH MORE EFFICIENT
        #     # i have a bug I don't get the result for 0x96c9ef
        # 
        #     # now just try another approach that is a swap matcher
        #     # in fact could run it on new cells or not properly found cells --> then it should not be so along and could even make it recursive...
        #     # --> think how to do that but should be ok
        #     print('swapped_cells', len(swapped_cells), swapped_cells)
    '''
    # do a matcher --> that gives the exchange to be done between first and second based on neighborhood

    # ça marche pas mais l'idee est bonne et c'est rapide --> trouver un moyen intelligent d'utiliser ça et de le fixer
    # what if there is a better match

    # inner_time = timer()
    best_pairs = {}  # dict that contains pairs -->  a good idea
    best_pairs_score = {}

    for key, value in t0_cells_n_vertices.items():
        max_match = 0.
        best_match = 0
        neighs_at_t0 = t0_cells_n_vertices[key]
        best_pair_size = 0

        # if key == 0x496bbd:
        #     print('pairs for a test', neighs_at_t0)

        for key2, value2 in t1_cells_n_vertices.items():
            neighs_at_t1 = t1_cells_n_vertices[key2]
            common_neighs = set(neighs_at_t1).intersection(neighs_at_t0)
            # if len(common_neighs) / ((len(neighs_at_t1)+len(neighs_at_t0))/2) > max_match:

            # if key2 == 0xd7bff1:
            #     print('rescue', neighs_at_t1)
            # if key2 ==

            # if two are equal --> try to take the best --> find a way to do it smartly
            if len(common_neighs) > 0 and len(
                    common_neighs) >= max_match:  # if found twice --> find a way to select which cells to take
                # max_match = len(common_neighs) / ((len(neighs_at_t1)+len(neighs_at_t0))/2)

                conti = True
                if len(common_neighs) == max_match:
                    if len(neighs_at_t1) == best_pair_size:
                        if key == key2:
                            # do favor identity as much as possible
                            conti = True
                        # find a trick or random
                        elif key2 in neighs_at_t0:  # not sure this is very smart but give it a try NB recursion may fix bugs
                            conti = False
                    elif len(neighs_at_t1) > best_pair_size:
                        conti = False
                    # si les 2 font la meme taille --> pb à nouveau --> y a t'il une astuce...

                if conti:
                    max_match = len(common_neighs)  # / ((len(neighs_at_t1)+len(neighs_at_t0))/2)
                    best_match = key2
                    best_pair_size = len(neighs_at_t1)
                    if __DEBUG__:
                        if key == cell_to_check:
                            hex_list_0 = [format(x, 'x') for x in neighs_at_t0]
                            hex_list_1 = [format(x, 'x') for x in neighs_at_t1]
                            print('pairs for a test', neighs_at_t0, neighs_at_t1, len(common_neighs), max_match,
                                  format(key, 'x'), format(key2, 'x'), key, key2, hex_list_0,
                                  hex_list_1)  # that works --> so indeed the best match is found but then cells are lost and swapped --> why is that --> I still miss a ctrl somewhere --> or a secondary swap that is less good somehow --> check that or a coding error!!
                else:
                    if __DEBUG__:
                        if key == cell_to_check:
                            hex_list_0 = [format(x, 'x') for x in neighs_at_t0]
                            hex_list_1 = [format(x, 'x') for x in neighs_at_t1]
                            print('discarded pairs for a test', neighs_at_t0, neighs_at_t1, len(common_neighs),
                                  max_match,
                                  format(key, 'x'), format(key2, 'x'), key, key2, hex_list_0,
                                  hex_list_1)
        if __DEBUG__:
            if key == cell_to_check:
                print('final match', max_match, cell_to_check, best_match, format(cell_to_check, 'x'),
                      format(best_match, 'x'))
        # or best --> compare to its current match and if better then take it otherwise ignore  --> maybe

        # in fact I take all here but is that really worth ??? --> maybe take one but only if the cell has just one neighbor ???? --> shall I start with 2 ??? that would dramatically reduce the nb of cells I guess --> think about it
        if max_match >= 0.5:  # take all cells that have at least half of their neighbs --> 40--> maybe ok
            # if key != best_match:  # only add if they are different and offer a swapping then always put so that people can see if better compared to orig

            # in fact need compare that there is really a gain otherwise forget about the swap --> check if the other cell is gaining --> check --> think how to do that --> try maybe for one cell just to know how this works!!!

            # check that there will be a gain otherwise skip...
            # print(t1_cells_n_vertices[key], 'vs', t1_cells_n_vertices[best_match])
            # comment faire je vois pas en fait...

            # compare neighbors of current match versus those in the previous
            best_pairs[key] = best_match
            best_pairs_score[key] = max_match  # will be used later to avoid issues


    # c'est moins bon qd je filtre ici on dirait et c'est presque aussi lent donc pas sur...
    # # maybe filter by score
    # print('best_pairs_score', len(best_pairs_score))
    # print('best_pairs', len(best_pairs))
    #
    # # TODO FIX AND CHECK THE LINE BELOW
    # MIN_CUT_OFF_SWAPPING = 2

    # remove cells that did not change much
    if MIN_CUT_OFF_SWAPPING is not None and MIN_CUT_OFF_SWAPPING>0:
        pairs_to_ignore = [k for k, v in best_pairs_score.items() if v <= MIN_CUT_OFF_SWAPPING]
        for ids in pairs_to_ignore:
            del best_pairs[ids]

        # print('best_pairs cleaned', len(best_pairs))

    # remove all entries with score less than

    # finding duplicate values
    # from dictionary using set
    rev_dict = {}
    for key, value in best_pairs.items():
        rev_dict.setdefault(value, set()).add(key)

    # from itertools import chain
    # result = set(chain.from_iterable(
    #     values for key, values in rev_dict.items()
    #     if len(values) > 1))
    # result = filter(lambda x: len(x) > 1, rev_dict.values())
    dupes_matches_to_be_removed = []
    for key, value in rev_dict.items():

        if __DEBUG__:
            if key == cell_to_check:
                print('MEGA THERE######')

            # if value == best_pairs[cell_to_check]:
            #     print('MEGA THERE2######\ndupe entry', rev_dict[value])

            # favor self here

        if len(value) > 1:
            # if key in value:
            #     best_match=key

            best_match = 0
            max_match = 0
            # keep only best and remove all others from the table
            for iii, val in enumerate(value):
                if __DEBUG__:
                    if key == cell_to_check:
                        print(iii, val, best_pairs_score[val])

                if best_pairs_score[val] == max_match:
                    if val == key:
                        max_match = best_pairs_score[val]
                        best_match = val
                elif best_pairs_score[val] > max_match:
                    max_match = best_pairs_score[val]
                    best_match = val

            if __DEBUG__:
                if key == cell_to_check:
                    print(value, val, max_match)

            # I have a bug here of a swapping with the leg_cropped images but fix it later cause still all is quite good --> why not detected I don't get it
            # if 0xfc34b0 in rev_dict:
            #     print('zoubi',rev_dict[0xfc34b0], best_match, value) # --> in one case need remove and not in the other
            # elif 0x29d785 in rev_dict:
            #     print('zouba',rev_dict[0x29d785], best_match, value)

            value.remove(best_match)
            # I do have a bug here but I don't get it in fact but ok for now
            dupes_matches_to_be_removed.extend(
                value)  # not a good idea --> just need to remove it from the value not from everywhere... is that correct ???? remove it only if present again in duplicates in final stuff after cleaning

    # for key, value in rev_dict.items():
    #     if len(value) > 1:
    #         print('final cleaning')

    if __DEBUG__:
        print('dupes_matches_to_be_removed', dupes_matches_to_be_removed,
              [format(x, 'x') for x in dupes_matches_to_be_removed])

    for key in dupes_matches_to_be_removed:
        if key in best_pairs:
            del best_pairs[key]

    # TODO --> in all cases where there are more cells than one for an output --> just keep the best score --> remove all others from the stuff

    # print(rev_dict)
    # # printing result
    # print("resultant key", str(result))

    # find all dupes in the values of a dict

    # if two there are twice the same cell in the stuff --> just keep the best

    # or could check overlap to find best match and if in addition improves neighborhood then take it --> think about it and how to do that

    # can I do a swap then --> good idea
    # --> color in the second should be replaced by color in the first and reversed --> see how I can do that

    # is that the swap --> did I do a mistake ???
    # check that the new assignment is better than the old one
    # maybe it's even a better way to do swap than what I used to do in my old swap code --> try it

    # TODO remove dupes --> keep best score

    # mapping is good but cells get swapped --> check the orange cells
    # maybe make this faster by coding this in C

    # about twice faster when not using np.where --> good
    def swap(pairs_of_cells_to_swap,
             tracked_cells_t1):  # this is super slow and would benefit from recoding it in C --> this takes most of the time of the function
        # inner_time = timer()
        clone_tracks = np.array(tracked_cells_t1, copy=True)
        # print('current time', timer() - inner_time)

        for cell1, cell2 in pairs_of_cells_to_swap.items():
            if cell1 == cell2: # remove identity cause it slows down a lot the process
                continue
            # coords1 = tracked_cells_t1[tracked_cells_t1 == cell1]
            # coords2 = tracked_cells_t1[tracked_cells_t1 == cell2]
            # coords1 = np.where(tracked_cells_t1 == cell1) # NB np.where is super slow --> never use it
            # coords2 = np.where(tracked_cells_t1 == cell2)

            # if cell2 not in pairs_of_cells_to_swap.keys(): # very slow --> find a way to make that faster
            clone_tracks[tracked_cells_t1 == cell1] = cell2
            clone_tracks[tracked_cells_t1 == cell2] = cell1

            if __DEBUG__:
                print('converting', cell2, '--> ', cell1, format(cell2, 'x'), '-->', format(cell1, 'x'))

            # find a way todo things and to not duplicate stuff

            # pb --> this creates duplications but otherwise it does swap correct cells --> find a way to prevent this --> I need to order swaps and other stuff
        # print('current time2', timer() - inner_time)
        return clone_tracks

    # 579 ou 580 par lot --> est-ce qu'il flippe vraiment tt ça ??? --> si je pouvais réduire ça ça irait bcp plus vite
    # print('pairs_of_cells_to_swap', len(best_pairs)) # that is really a lot of things to swap --> do I really need to do that ???
    # check the ones that are really worth swapping I guess they must have a cutoff

    # inner_time = timer()
    tracked_cells_t1 = swap(best_pairs,
                            tracked_cells_t1)  # c'est ce truc qui est super lent --> plus de la moitié du temps --> try improve that
    # print('current time2', timer() - inner_time)

    # then try recolor all of the cells

    # nb if there are dupes --> need take the best score in fact --> TODO

    if __DEBUG__:
        print('best_pairs', len(best_pairs), best_pairs)  # maybe best ratio versus its own size
    # TODO maybe save swapped cells --> only show those and blacken all the rest

    # print('final pair test', best_pairs[0xfea247])
    # print(0xfea247, 0x961f81)
    return tracked_cells_t1


if __name__ == '__main__':

    swap_fix_recursions = 3

    # pb is because after translation the cell is also present in both --> skipped by my stuff --> need detect those cells and chose to get only one in fact
    # pb is because after translation the cell is also present in both --> skipped by my stuff --> need detect those cells and chose to get only one in fact
    # or first do the

    start_all = timer()

    # path = '/E/Sample_images/tracking_test/leg_uncropped'  # quick test of the tracking algo
    # path = '/E/Sample_images/tracking_test/leg_cropped'  # quick test of the tracking algo
    path = '/E/Sample_images/tracking_test/test_uncropped'  # quick test of the tracking algo
    # path = '/E/Sample_images/tracking_test/test_cropped'  # quick test of the tracking algo

    if not path.lower().endswith('.txt') and not path.lower().endswith('.lst'):
        images_to_analyze = os.listdir(path)
        images_to_analyze = [os.path.join(path, f) for f in images_to_analyze if
                             os.path.isfile(os.path.join(path, f))]  # list only files and only if they end by tif
        images_to_analyze = natsorted(images_to_analyze)
    else:
        images_to_analyze = loadlist(path)

    # get current and next in fact and map next

    # process first image directly then move on to next
    # images_to_analyze = []

    # for image 0 --> just randomly fill it with colors --> make sure no dupe and that's it

    # TODO --> do the first

    print(images_to_analyze)

    for l in range(len(images_to_analyze) - 1):

        start_loop = timer()

        file_path_0 = images_to_analyze[l]
        file_path_1 = images_to_analyze[l + 1]

        print('files', file_path_1, file_path_0)

        if not file_path_1.endswith('.tif') or not file_path_0.endswith('.tif'):
            continue

        # print('files', file_path_1, file_path_0)

        filename0_without_ext = os.path.splitext(file_path_0)[0]
        filename1_without_ext = os.path.splitext(file_path_1)[0]

        tracked_cells_t0 = Img(os.path.join(filename0_without_ext, 'tracked_cells_resized.tif'))
        tracked_cells_t1 = Img(os.path.join(filename1_without_ext, 'tracked_cells_resized.tif'))

    vertices_t0 = detect_vertices_and_bonds(RGB_to_int24(tracked_cells_t0))
    vertices_t0 = np.where(vertices_t0==255)
    vertices_t1 = detect_vertices_and_bonds(RGB_to_int24(tracked_cells_t1))
    vertices_t1 = np.where(vertices_t1 == 255)

    # time is always increasing but why is that is some set not reset
    for fix in range(swap_fix_recursions):
        print('swapping fix recursion #', fix)
        begin_loop = timer()
        tracked_cells_t1 = _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1, vertices_t0=vertices_t0, vertices_t1=vertices_t1)
        # tracked_cells_t1 = _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1)
        print('current time', timer() - begin_loop)
        # Img(tracked_cells_t1).save(os.path.join(filename0_without_ext, 'track_after_swap_correction'+ str(fix)+ '.tif'), mode='raw')

    # example of swapped cells
    print('total time', timer() - start_all)  # --> 47 secs --> un peu lent qd meme

    Img(int24_to_RGB(tracked_cells_t1)).save(os.path.join(filename0_without_ext, 'track_after_swap_correction.tif'), mode='raw')

    # puis je la faire dynamiquement pr eviter les swaps en live
    # --> void aussi si rentable ou pas

    # TODO -->
    # this now seems ok and could be done recursively...
    # could also use neighbor cells to reassign most likely cell --> cell that matches most of the surrounding IDs

    # pas mal --> see how many recursions might be needed to get the stuff to work --> might be time consuming to do plenty of recursion... --> maybe just update the map so that not everything needs be calculated again

    # if a cell is unsure try reassign it by its neighborood...

    # TODO --> reidentify cells by neighborhood --> simply use the vertices to detect the neighbors --> split the code



import traceback
import matplotlib.pyplot as plt

from epyseg.ta.measurements.measurements3D.get_point_on_surface_if_centroid_is_bad import point_on_surface
from epyseg.ta.utils.TA_tools import get_TA_clone_file_name
from epyseg.img import RGB_to_int24, Img
from epyseg.ta.tracking.tools import smart_name_parser
from skimage.measure import label, regionprops
import numpy as np
from epyseg.tools.early_stopper_class import early_stop
from epyseg.tools.logger import TA_logger  # logging

logger = TA_logger()


# TODO --> I would really need an editor --> see how I can do that ???
# because that would really be super useful !!!

# this class will gather all the TA selection tools and conversion between the various formats

# if not return image then return
# converts a set of coordinates to a list of Ids or to an image output --> will be very useful to plot and store clones, dividing cells and dying cells...

# deprecated --> use another code if possible
def convert_db_xy_coordiantes_to_local_cell_ID_for_clones(db_path, table_name, image_for_correspondance_label_or_RGB24,
                                                          return_image=False):
    from epyseg.ta.database.sql import TAsql
    # read the coords from the table and make sure they can be used to generate a file
    # pass

    # open the db and get the coords and return a set of cells or an image depending on what the user wants to have -> maybe could also return a table or a temp table so that one can do a join on it --> very smart idea too!!
    output = []
    db = None
    try:

        db = TAsql(db_path)
        # now i need get the coords from the db and check the image for that

        # NB THIS TABLES ASSUMES THE QUERY RETURNS 2D COORDS --> ALWAYS AN X AND A Y
        #headers,
        query_results = db.run_SQL_command_and_get_results('SELECT * FROM ' + table_name, return_header=False)

        # formatted = np.asarray(query_results, dtype=object)
        # print(headers)
        # print(query_results)
        # print(formatted)

        # display = None

        if return_image:
            output = np.zeros_like(image_for_correspondance_label_or_RGB24)

        # TODO --> need a check to be sure the coords are really inside the cell !!! --> normally it must always be but that is a smart check anyway!!!
        for coords in query_results:
            # print(coords)
            # print(coords[0], coords[1])
            color = image_for_correspondance_label_or_RGB24[coords[0], coords[1]]
            # this gives me all the cells I need to get
            # then I could either get the corresponding label or the image corresponding to that
            # print(color)
            if return_image:
                output[image_for_correspondance_label_or_RGB24 == color] = color
            else:
                output.append(color)
    except:
        traceback.print_exc()
        logger.error('Something went wrong, selection could not be converted to 2D coordinates')
    finally:
        if db is not None:
            try:
                db.close()
            except:
                traceback.print_exc()
        return output




def get_colors_drawn_over(mask, image_to_analyze, forbidden_colors=[0,0xFFFFFF]):
    # print('I was called')
    if mask is None:
        return
    # print(mask)
    # need get all the colors of the raw image that correspond to user selection
    selected_colors = image_to_analyze[mask != 0]
    # print('selected_colors',selected_colors)

    # why is the user drawing null
    # plt.imshow(mask)
    # plt.show()

    if len(image_to_analyze.shape)==3:
        selected_colors = RGB_to_int24(selected_colors)
    selected_colors = set(selected_colors.ravel().tolist())


    # if 0xFFFFFF in selected_colors:
    #     selected_colors.remove(0xFFFFFF)
    if forbidden_colors is not None:
        if not isinstance(forbidden_colors, list):
            forbidden_colors = [forbidden_colors]
        for color in forbidden_colors:
            if color in selected_colors:
                selected_colors.remove(color)
    selected_colors = list(selected_colors)

    # print(selected_colors) # all is ok there -> so the error is afterwards
    return selected_colors

def handCorrection_to_label(handCorrection):
    if isinstance(handCorrection,str):
        # TODO shall I so more complex things such as try also create the handcorrection name if it does not exist ??? --> TODO but maybe later!!!
        handCorrection = Img(handCorrection)
    if len(handCorrection.shape) == 3:
        handCorrection= handCorrection[...,0]
    handCorrection = label(handCorrection, connectivity=1, background=255)
    return handCorrection

def add_coords_to_db(db_path, table_name_to_store_coords_to, coords):
    from epyseg.ta.database.sql import TAsql
    coords_of_selection = {'pixel_within_cell_y': [],
                           'pixel_within_cell_x': [],
                           }
    db = None
    try:
        db = TAsql(db_path)
        for coord in coords:
            coords_of_selection['pixel_within_cell_y'].append(coord[0])
            coords_of_selection['pixel_within_cell_x'].append(coord[1])
        db.create_and_append_table(table_name=table_name_to_store_coords_to, datas=coords_of_selection)
    except:
        traceback.print_exc()
        logger.error('something went wrong while trying to write coordinates to the DB')
    finally:
        if db is not None:
            try:
                db.close()
            except:
                traceback.print_exc()

# img_to_analyze
def convert_selection_color_to_coords(img_to_analyze_RGB24_or_label, selected_cells, bg_color=None):
    # tmp = img_to_analyze
    coords_of_selection = []
    if img_to_analyze_RGB24_or_label is None:
        logger.error('no image -> nothing to do...')
        return coords_of_selection

    if len(img_to_analyze_RGB24_or_label.shape) == 3:
        img_to_analyze_RGB24_or_label = RGB_to_int24(img_to_analyze_RGB24_or_label)


    if bg_color is None:
        if 0xFFFFFF in img_to_analyze_RGB24_or_label:
            bg_color = 0xFFFFFF
        else:
            bg_color = 0

    cell_label = label(img_to_analyze_RGB24_or_label, connectivity=1, background=bg_color)

    # for all the selected cells --> return the coords --> probably a quite good idea...
    for region in regionprops(cell_label):
        color = img_to_analyze_RGB24_or_label[region.coords[0][0], region.coords[0][1]]
        if color in selected_cells:
            point_awlays_inside_the_cell = point_on_surface(region, cell_label)
            coords_of_selection.append(point_awlays_inside_the_cell)

    return coords_of_selection

# def convert_coords_to_image(image_for_correspondance_label_or_RGB24, coords):
#     for coord in coords:
#             # print(coords)
#             # print(coords[0], coords[1])
#             color = image_for_correspondance_label_or_RGB24[coord[0], coord[1]]
#             # this gives me all the cells I need to get
#             # then I could either get the corresponding label or the image corresponding to that
#             # print(color)
#             if return_image:
#                 output[image_for_correspondance_label_or_RGB24 == color] = color
#             else:
#                 output.append(color)

# converts back some local coords to IDs
def convert_coords_to_IDs(img_to_analyze_RGB24_or_label, selected_coords, forbidden_colors = [0xFFFFFF,0], return_image=False, new_color_to_give_to_cells_if_return_image=None):
    ids = []
    if img_to_analyze_RGB24_or_label is None:
        logger.error('Empty image --> nothing to do...')
        if return_image:
            return None
        return ids

    if len(img_to_analyze_RGB24_or_label.shape) == 3:
        img_to_analyze_RGB24_or_label = RGB_to_int24(img_to_analyze_RGB24_or_label)
    if return_image:
        ids = np.zeros_like(img_to_analyze_RGB24_or_label)
    for coords in selected_coords:
        color = img_to_analyze_RGB24_or_label[coords[0], coords[1]]
        if forbidden_colors is not None:
            if color in forbidden_colors:
                continue
        if not return_image:
            ids.append(color)
        else:
            # can be used to plot clones
            if new_color_to_give_to_cells_if_return_image is not None:
                ids[img_to_analyze_RGB24_or_label == color] = new_color_to_give_to_cells_if_return_image
            else:
                ids[img_to_analyze_RGB24_or_label==color]=color
    return ids

# also do the oposite --> convert cells to database coords
# maybe I need ask the color for the conversion --> that may make sense or assume balck or white by default --> think about it ???? -->
# otherwise I also need the db_path
# and the table name
# depreacated --> use another code if possible...
def convert_selected_cells_to_local_db_coords(selected_cells, label_or_RGB24_image_corresponding_to_selected_cells,
                                              db_path, table_name_to_store_coords_to, bg_color_for_label=None):
    if not selected_cells:
        logger.error('Nothing selected --> nothing to do')
        return

    if bg_color_for_label is None:
        # auto mode to determine the label --> may fail if the image is non conventional
        if 0xFFFFFF in label_or_RGB24_image_corresponding_to_selected_cells:
            bg_color_for_label = 0xFFFFFF
        else:
            bg_color_for_label = 0

    # db = None
    try:
        cell_label = label(label_or_RGB24_image_corresponding_to_selected_cells, connectivity=1,
                           background=bg_color_for_label)
        # rps =

        # check if region props contains a point
        # or get the color of the point and return the corresponding cell --> TODO
        # get the index of the

        # cells_2D['pixel_within_cell_y'].append(point_awlays_inside_the_cell[0])
        # cells_2D['pixel_within_cell_x'].append(point_awlays_inside_the_cell[1])

        # need create a db with the right keys --> TODO

        # coords_of_selection = {'pixel_within_cell_y': [],
        #                        'pixel_within_cell_x': [],
        #                        }

        # db = TAsql(db_path)
        # print('cells_2D'*20)
        coords = []
        for region in regionprops(cell_label):
            color = label_or_RGB24_image_corresponding_to_selected_cells[region.coords[0][0], region.coords[0][1]]
            if color in selected_cells:
                # get the coord always inside the cell and store it to a db --> TODO
                # can also be used for a reload --> very smart in fact!!!!
                point_awlays_inside_the_cell = point_on_surface(region, cell_label)
                coords.append(point_awlays_inside_the_cell)
                # print(point_awlays_inside_the_cell)
                # coords_of_selection['pixel_within_cell_y'].append(point_awlays_inside_the_cell[0])
                # coords_of_selection['pixel_within_cell_x'].append(point_awlays_inside_the_cell[1])

        # db.create_and_append_table(table_name=table_name_to_store_coords_to, datas=coords_of_selection)
        add_coords_to_db(db_path, table_name_to_store_coords_to, coords)

    except:
        traceback.print_exc()
        logger.error('Something went wrong, selection could not be converted to 2D coordinates')
    # finally:
    #     if db is not None:
    #         try:
    #             db.close()
    #         except:
    #             traceback.print_exc()

    # gets an image with some selected cells and convert it to a set of local coords that can then be shown and or updated
    # pass

# maybe store the ids of the cells here then!!!
# TODO --> see my other code and get it there
# faire differents selecteurs et les mettre à tel ou tel endroit --> TODO
# TODO hack ImgDisplayWindow.display to get the drawing of the user and the selection --> TODO

if __name__ == '__main__':
    import sys

    if False:
        # test how to append selection to a clone database
        path_to_image = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_empty/focused_Series012.png'
        selected_cells = [200, 201, 302]  # cells 1-3 have been selected
        handCorrection = Img(smart_name_parser(path_to_image, ordered_output='handCorrection.tif'))
        if len(handCorrection.shape) == 3:
            handCorrection = handCorrection[..., 0]
        cell_label = label(handCorrection, connectivity=1, background=255)
        # print(smart_name_parser(path_to_image, ordered_output='pyTA.db'))
        convert_selected_cells_to_local_db_coords(selected_cells, cell_label,
                                                  smart_name_parser(path_to_image, ordered_output='pyTA.db'),
                                                  'test_db_name', bg_color_for_label=0)  # smart
        # try relaod the coords as an image
        sys.exit(0)

    # ça marche vraiment bien en fait j'adore!!!
    if True:
        # load database from image and get the corresponding cell image or db depending on what is desired --> TODO
        path_to_image = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_empty/focused_Series012.png'
        handCorrection = Img(smart_name_parser(path_to_image, ordered_output='handCorrection.tif'))
        if len(handCorrection.shape) == 3:
            handCorrection = handCorrection[..., 0]
        cell_label = label(handCorrection, connectivity=1, background=255)
        output = convert_db_xy_coordiantes_to_local_cell_ID_for_clones(smart_name_parser(path_to_image, ordered_output='pyTA.db'), 'test_db_name', cell_label, return_image=True) #return_image=Fals
        # ça marche vraiment super bien
        if isinstance(output, np.ndarray):
            plt.imshow(output)
            plt.show()
        else:
            print(output)
        # --> this does return cells 1 to 3
        sys.exit(0)



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import sys
from qtpy.QtWidgets import QApplication, QFileDialog
from os.path import expanduser
import os


# class Open_Save_dialogs():

def openFileNameDialog(parent_window=None, extensions="Supported Files (*.jpg *.tif *.png);;All Files (*)",
                       path=expanduser('~')):
    # options = QFileDialog.Options()
    # options |= QFileDialog.DontUseNativeDialog
    # QFileDialog.setOption(QFileDialog.DontUseNativeDialog, True)
    if not os.path.exists(path):
        path = expanduser('~')
    fileName, _ = QFileDialog.getOpenFileName(parent_window, "Select a File", path, extensions, options=QFileDialog.DontUseNativeDialog)
    return fileName

def openFileNamesDialog(parent_window=None, extensions="Supported Files (*.jpg *.tif *.png);;All Files (*)",
                        path=expanduser('~')):
    # options = QFileDialog.Options()
    # options |= QFileDialog.DontUseNativeDialog
    if not os.path.exists(path):
        path = expanduser('~')
    files, _ = QFileDialog.getOpenFileNames(parent_window, "Select Files", path, extensions, options=QFileDialog.DontUseNativeDialog)
    return files

# not so easy to do in fact
# def openSingleFileOrDirectoryDialog(self, parent_window=None, extensions="Supported Files (*.jpg *.tif *.png);;All Files (*)",
#                        path=expanduser('~')):
#     setFileMode(QFileDialog::Directory | QFileDialog::ExistingFiles)
#     options = QFileDialog.Options()
#     options |= QFileDialog.DontUseNativeDialog
#     if not os.path.exists(path):
#         path = expanduser('~')
#     fileName, _ = QFileDialog.getOpenFileName(parent_window, "QFileDialog.getOpenFileName()", path,
#                                               extensions, options=options)


# create file if does not exist and delete if name varies --> maybe do this as an option...
def saveFileDialog(parent_window=None, path=expanduser('~'),  extensions="All Files (*);;Text Files (*.txt)", default_ext=None):
    # options = QFileDialog.Options()
    # options |= QFileDialog.DontUseNativeDialog
    try:
        # nice hack to allow to specify a default name even if the file does not exist as long as the parent folder exists
        if not os.path.exists(path) and not os.path.exists(os.path.dirname(path)):
            path = expanduser('~')
    except:
        path = expanduser('~')
    fd = QFileDialog(parent_window, "Save a File", path, extensions, options=QFileDialog.DontUseNativeDialog)
    if default_ext is not None:
        fd.setDefaultSuffix(default_ext)
    fd.setAcceptMode(QFileDialog.AcceptSave)
    selected = fd.exec()
    if selected:
        fileName = fd.selectedFiles()[0]
        return fileName
    else:
        return

def openDirectoryDialog(parent_window=None, path=expanduser('~')):
    # options = QFileDialog.Options()
    # options |= QFileDialog.DontUseNativeDialog
    if not os.path.exists(path):
        path = expanduser('~')
    folderName = QFileDialog.getExistingDirectory(parent_window, "Select a Directory", path, options= QFileDialog.DontUseNativeDialog)
    if folderName is not None:
        if not folderName.strip():
            return None
        if not folderName.endswith("/") and not folderName.endswith("\\"):
            folderName += '/'
    return folderName

if __name__ == '__main__':
    app = QApplication(sys.argv)
    # ex = Open_Save_dialogs()
    filename = saveFileDialog(default_ext='.tif')
    print(filename)
    sys.exit(0)


import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import sys



from qtpy.QtCore import QRect, QTimer
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QWidget

from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.draw.shapes.square2d import Square2D
from epyseg.draw.widgets.vectorial import VectorialDrawPane
from qtpy.QtWidgets import QMenu, QApplication
from qtpy import QtCore, QtGui

from epyseg.img import toQimage
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

class Createpaintwidget(QWidget):

    def __init__(self):
        super().__init__()
        self.vdp = VectorialDrawPane(active=False) #, demo=True
        self.image = None
        self.imageDraw = None
        self.cursor = None
        self.maskVisible = True
        self.scale = 1.0
        self.drawing = False
        self.brushSize = 3
        self._clear_size = 30
        self.drawColor = QtGui.QColor(QtCore.Qt.red) # blue green cyan
        self.eraseColor = QtGui.QColor(QtCore.Qt.black)
        self.cursorColor = QtGui.QColor(QtCore.Qt.green)
        self.lastPoint = QtCore.QPoint()
        self.change = False
        # KEEP IMPORTANT required to track mouse even when not clicked
        self.setMouseTracking(True)  # KEEP IMPORTANT
        self.scrollArea = None
        self.statusBar = None

    def setImage(self, img):
        if img is None:
            self.image = None
            self.imageDraw = None
            self.update()
            return
        else:
            self.image = toQimage(img) #.getQimage() # bug is here

            # self.image = QPixmap(100,200).toImage()
        width = self.image.size().width()
        height = self.image.size().height()
        top = self.geometry().x()
        left = self.geometry().y()
        self.setGeometry(top, left, width*self.scale, height*self.scale)
        self.imageDraw = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
        self.imageDraw.fill(QtCore.Qt.transparent)
        self.cursor = QtGui.QImage(self.image.size(), QtGui.QImage.Format_ARGB32)
        self.cursor.fill(QtCore.Qt.transparent)
        self.update()

    def mousePressEvent(self, event):
        if not self.hasMouseTracking():
            return
        self.clickCount = 1
        if self.vdp.active:
            self.vdp.mousePressEvent(event)
            self.update()
            return

        if event.buttons() == QtCore.Qt.LeftButton or event.buttons() == QtCore.Qt.RightButton:
            self.drawing = True
            zoom_corrected_pos = event.pos() / self.scale
            self.lastPoint = zoom_corrected_pos
            self.drawOnImage(event)

    def mouseMoveEvent(self, event):
        if not self.hasMouseTracking():
            return
        # print('in mouse move', self.hasMouseTracking(), self.drawing, self.vdp.active)
        if self.statusBar:
            zoom_corrected_pos = event.pos() / self.scale
            self.statusBar.showMessage('x=' + str(zoom_corrected_pos.x()) + ' y=' + str(
                zoom_corrected_pos.y()))
        if self.vdp.active:
            self.vdp.mouseMoveEvent(event)
            region = self.scrollArea.widget().visibleRegion()
            self.update(region)
            return
        self.drawOnImage(event)

    def drawOnImage(self, event):
        zoom_corrected_pos = event.pos() / self.scale
        if self.drawing and (event.buttons() == QtCore.Qt.LeftButton or event.buttons() == QtCore.Qt.RightButton):
            # now drawing or erasing over the image
            painter = QtGui.QPainter(self.imageDraw)
            if event.buttons() == QtCore.Qt.LeftButton:
                painter.setPen(QtGui.QPen(self.drawColor, self.brushSize, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                          QtCore.Qt.RoundJoin))
            else:
                painter.setPen(QtGui.QPen(self.eraseColor, self.brushSize, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                          QtCore.Qt.RoundJoin))
            if self.lastPoint != zoom_corrected_pos:
                painter.drawLine(self.lastPoint, zoom_corrected_pos)
            else:
                # if zero length line then draw point instead
                painter.drawPoint(zoom_corrected_pos)
            painter.end()

        # Drawing the cursor TODO add boolean to ask if drawing cursor should be shown
        painter = QtGui.QPainter(self.cursor)
        # We erase previous pointer
        r = QtCore.QRect(QtCore.QPoint(), self._clear_size * QtCore.QSize() * self.brushSize)
        painter.save()
        r.moveCenter(self.lastPoint)
        painter.setCompositionMode(QtGui.QPainter.CompositionMode_Clear)
        painter.eraseRect(r)
        painter.restore()
        # draw the new one
        painter.setPen(QtGui.QPen(self.cursorColor, 2, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                  QtCore.Qt.RoundJoin))
        painter.drawEllipse(zoom_corrected_pos, self.brushSize / 2.,
                            self.brushSize / 2.)
        painter.end()
        region = self.scrollArea.widget().visibleRegion()
        self.update(region)

        # required to erase mouse pointer
        self.lastPoint = zoom_corrected_pos

    def mouseReleaseEvent(self, event):
        if not self.hasMouseTracking():
            return
        if self.vdp.active:
            self.vdp.mouseReleaseEvent(event)
            self.update()  # required to update drawing
            return
        if event.button == QtCore.Qt.LeftButton:
            self.drawing = False
        if self.clickCount == 1:
            QTimer.singleShot(QApplication.instance().doubleClickInterval(),
                              self.updateButtonCount)

    # adds context/right click menu but only in vectorial mode
    def contextMenuEvent(self, event):
        if not self.vdp.active:
            return
        cmenu = QMenu(self)
        newAct = cmenu.addAction("New")
        opnAct = cmenu.addAction("Open")
        quitAct = cmenu.addAction("Quit")
        action = cmenu.exec_(self.mapToGlobal(event.pos()))
        if action == quitAct:
            sys.exit(0)

    def updateButtonCount(self):
        self.clickCount = 1

    def mouseDoubleClickEvent(self, event):
        self.clickCount = 2
        self.vdp.mouseDoubleClickEvent(event)

    def paintEvent(self, event):
        canvasPainter = QtGui.QPainter(self)
        # the scrollpane visible region
        visibleRegion = self.scrollArea.widget().visibleRegion()
        # the corresponding rect
        visibleRect = visibleRegion.boundingRect()
        # the visibleRect taking zoom into account
        scaledVisibleRect = QRect(visibleRect.x() / self.scale, visibleRect.y() / self.scale,
                                  visibleRect.width() / self.scale, visibleRect.height() / self.scale)
        if self.image is None:
            canvasPainter.eraseRect(visibleRect)
            canvasPainter.end()
            return

        canvasPainter.drawImage(visibleRect, self.image, scaledVisibleRect)
        if not self.vdp.active and self.maskVisible:
            canvasPainter.drawImage(visibleRect, self.imageDraw, scaledVisibleRect)
            # should draw the cursor
        canvasPainter.drawImage(visibleRect, self.cursor, scaledVisibleRect)

        if self.vdp.active:
            self.vdp.paintEvent(canvasPainter, scaledVisibleRect)
        canvasPainter.end()
